/** \file   dif1d.cpp
    \brief  provides analytical solution to 1d diffusion problem in a cylindrical geometry with inner sink
    \author EV, KL
    \date   2011-2015
    \version 1.2

This program computes the flux in one-dimensional diffusion problem in a cylindrical geometry, 
in application to studying the loss-cone problem of consumption of stars by a supermassive 
black hole.
The evolution of the distribution function \f$ f(R,t) \f$ in the interval 
\f$ R_{min} <= R <= 1 \f$ is governed by the equation:
\f[
  \frac{ \partial f(R,t) }{ \partial t} = 
  \frac{\partial}{\partial R} \left( D\, R\, \frac{\partial f(R,t)}{\partial R} \right) ,
\f]
with initial conditions \f[
  f(R,0) = \left[ \begin{array}{ll} (R/R_\mathrm{gap})^p &, R<R_\mathrm{gap} \\ 
    1 &, R\ge R_\mathrm{gap}  \end{array} \right. ,
\f], 
subject to the boundary conditions 
\f[
  f(R_{min},t) = \alpha\, R_{min}\, \partial f/\partial R ; \quad
  \frac{\partial}{\partial R} f(1,t) = 0 .
\f]

The general form of the boundary condition interpolates between empty 
(\f$ \alpha \ll 1 \f$) and full (\f$ \alpha \gg 1 \f$) regimes.
The solution for the case \f$ \alpha=0 \f$ is given by [Milosavljevic&Merritt 2003] in terms of 
series of Bessel functions (MM03, Eqs.24-28). 
The generalization to arbitrary \f$ \alpha \f$ is done by modifying the expression for the roots 
\f$ \beta_m \f$ of Bessel functions.

The series solution is truncated at a finite number of terms MMAX; 
it does not provide an accurate solution at small times, so we use a numerical solution 
obtained with a finite-difference scheme for t<tcutoff.
The cutoff time is computed from the requirement that 
a) the exponential decay of highest-m terms in the series expansion is at least exp(-30)=1e-13,
b) the flux(t) and cumulative flux(t) are not negative and not dominated by roundoff errors.

The initial conditions (IC) are specified in a quite flexible form; setting Rgap=0 gives isotropic IC 
(f(R)=1), setting Rgap=1 and p>0 gives tangentially-anisotropic IC, and setting Rmin<Rgap gives IC 
with a gap at low angular momenta, which can result from the ejection of stars by a binary black hole.

Version history:
1.0 - first release, for the paper [Vasiliev&Merritt 2013, ApJ]
1.1 - corrected computation of draining flux, for the paper [Vasiliev 2014, CQG]
1.2 - eliminated the computation of draining flux and switched to a more general
      Robin-type boundary conditions and more general initial conditions (with a gap),
      for the paper [Lezhnin&Vasiliev 2015]

*/

#include <fstream>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstddef>
#include <cmath>
#include <vector>
#include <iostream>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_linalg.h>
#include "config.h"
double gsl_pow_2(const double x) { return x*x; }

// global variables that are initialized in the main function and are used in other fncs
// parameters that are initialized from the command line
double D=1;               // diffusion coefficient D. Note that in MM03 a quantity mu is used (mu=D/4)
double alpha=0;           // coefficient in the boundary condition (<<1 -- empty, >>1 -- full loss cone regimes)
double Rmin=0;            // capture boundary (squared normalized angular momentum of the capture orbit)
double Rgap=0;            // width of gap region
double Pgap=0;            // power-law index of f(R)=(R/Rgap)^p in the gap region, 0 means f=0
double tmax=0;            // total time interval for the solution
double tstep=0;           // timestep at which the output is displayed
// other variables computed from the above parameters
double tcutoff=0;         // the cutoff time, flux on times less than cutoff time is computed numerically
double lc0=0;             // =sqrt{Rmin}
int MMAX=100;             // number of Bessel functions in the expansion of the initial d.f. (m=1..MMAX)
std::vector<double> beta; // roots of Bessel functions
std::vector<double> J0low, J1low, J1upp, Y1low, Y1upp;  // values of Bessel fncs at lower and upper ends of interval
std::vector<double> coef; // coefficients of the expansion


// a replacement for the default error handler that just displays error message and doesn't crash program
static void gsl_verbose_error_handler (const char *reason, const char *file, int line, int gsl_errno)
{
    if(gsl_errno == GSL_ETOL || gsl_errno == GSL_EROUND) return;  // don't bother
    std::cerr << "GSL error "<<gsl_errno<<" in "<<file<<" at line "<<line<<": "<<reason<<"\n";
}

double InitialDistribution(double x)
{
    if(Rgap<=0) return 1;
    if(Pgap==0) return x*x<Rgap ? 0 : 1;
    return std::min<double>(pow(x*x/Rgap, Pgap), 1);
}

// finite-difference solver, used for computing the solution at small "t".
// it operates with the equation in (R,t) space, where R=j^2, 
// and uses a fully implicit flux-conservative numerical integration scheme
gsl_vector *v_abovediag, *v_belowdiag, *v_diag, *v_rhs, *v_sol;  // used in tridiagonal linear eq.system solver
int ngrid;   // number of grid cells
std::vector<double> gridR;  // size: ngrid+1, edges of grid cells
std::vector<double> c_minus, c_plus;  // coefs of tridiagonal matrix, size: ngrid-1
double c_plus0, c_diag0;    // two additional coefs for the boundary condition at Rmin
double solverTime=0, solverMaxStep=0, initialSumF=0;

// finite-difference approximation to the integral of f over the entire domain
double solverSumF()
{
    double sum=0;
    for(int k=0; k<ngrid; k++)
        sum+=gsl_vector_get(v_sol, k)*(gridR[k+1]-gridR[k]);
    return sum;
}

// set up grid, allocate arrays and compute finite-difference coefficients
void prepareNumericSolution()
{
    ngrid = (int)(log(1/Rmin)*10);
    gridR.assign(ngrid+1, 0);
    for(int k=0; k<=ngrid; k++)
        gridR[k] = pow(Rmin, 1.0-k*1.0/ngrid);
    c_minus.assign(ngrid-1, 0);
    c_plus .assign(ngrid-1, 0);
    v_abovediag=gsl_vector_alloc(ngrid-1);
    v_belowdiag=gsl_vector_alloc(ngrid-1);
    v_diag=gsl_vector_alloc(ngrid);
    v_sol =gsl_vector_alloc(ngrid);
    v_rhs =gsl_vector_alloc(ngrid);
    c_plus0=D/(gridR[2]-gridR[0]);
    c_diag0=D*(gridR[1]+gridR[2]-2*gridR[0])/(gridR[2]-gridR[0])/(gridR[1]-gridR[0]);
    for(int k=0; k<ngrid; k++) {
        if(k>0)
            c_minus[k-1] = gridR[k  ]*D/(gridR[k+1]-gridR[k])/(gridR[k+1]-gridR[k-1])*2;
        if(k<ngrid-1)
            c_plus [k  ] = gridR[k+1]*D/(gridR[k+1]-gridR[k])/(gridR[k+2]-gridR[k  ])*2;
        gsl_vector_set(v_sol, k, InitialDistribution( sqrt((gridR[k+1]+gridR[k])/2) ));
    }
    // detemine "safe" timestep
    solverMaxStep=std::min<double>(1e-6/D, tmax/100);
    initialSumF=solverSumF();
}

// advance the solution by a timestep dt
void solverInternalStep(double dt)
{
    for(int k=0; k<ngrid-1; k++) {
        gsl_vector_set(v_abovediag, k, -c_plus [k]*dt);
        gsl_vector_set(v_belowdiag, k, -c_minus[k]*dt);
    }
    for(int k=0; k<ngrid; k++) {
        gsl_vector_set(v_diag, k, 1.0 + (k<ngrid-1?c_plus[k]*dt:0) + (k>0?c_minus[k-1]*dt:0));
        gsl_vector_set(v_rhs, k, gsl_vector_get(v_sol, k));
    }
    // boundary condition at Rmin
    gsl_vector_set(v_rhs,  0, alpha*gsl_vector_get(v_rhs, 0));
    gsl_vector_set(v_diag, 0, alpha*gsl_vector_get(v_diag, 0) + c_diag0*dt);
    gsl_vector_set(v_abovediag, 0, alpha*gsl_vector_get(v_abovediag, 0) - c_plus0*dt);
    // solve the linear system
    gsl_linalg_solve_tridiag(v_diag, v_abovediag, v_belowdiag, v_rhs, v_sol);
    solverTime+=dt;
}

// advance the solution by an arbitrary large timestep, breaking it down into smaller ones if necessary
void solverStep(double dt)
{
    while(dt>0) {
        solverInternalStep(std::min<double>(dt, solverMaxStep));
        dt-=solverMaxStep;
    }
}

// compute the flux through R=Rmin boundary from the numerical solution
double solverFlux()
{
    return (gsl_vector_get(v_sol, 1)-gsl_vector_get(v_sol, 0))/(gridR[2]-gridR[0])*2*gridR[1]*D;
}

// compute total flux through R=Rmin boundary
double solverFluxTotal()
{
    return initialSumF-solverSumF();
}

// analytical solution in terms of series expansion
// function to integrate in (25) MM03
struct intparam {
    double beta;
    double J1beta;
    double Y1beta;
    double argadd;
};

// the product A(x,y)*f(y)*y that enters the expression for the expansion coefficients (Eq.24 in MM03)
double fncAymultInitialDistribution(double y, void* params)
{
    double beta=((intparam*)params)->beta;
    return (gsl_sf_bessel_J0(beta*y)*((intparam*)params)->Y1beta - 
            gsl_sf_bessel_Y0(beta*y)*((intparam*)params)->J1beta) * y * InitialDistribution(y);
}

// same as above, but with the oscillating terms subtracted (using an asymptotic approximation for Bessel functions)
double fncAymultInitialDistribution_nonOsc(double y, void* params)
{
    double betay=((intparam*)params)->beta*y;
    double sq=sqrt(2./M_PI/betay);
    return ( (gsl_sf_bessel_J0(betay)-sq*cos(betay-M_PI/4) )*((intparam*)params)->Y1beta - 
             (gsl_sf_bessel_Y0(betay)-sq*sin(betay-M_PI/4) )*((intparam*)params)->J1beta ) * y * InitialDistribution(y);
}

// oscillating terms extracted from the integrand A(x,y)*f(y)*y
double fncAymultInitialDistribution_Osc(double y, void* params)
{
    double yy=y+((intparam*)params)->argadd;
    double sq=sqrt(2./M_PI/((intparam*)params)->beta/yy);
    return sq * yy * InitialDistribution(yy);
}

// compute m-th term in series expansion, by integrating separately the oscillatory and non-oscillatory parts of A(x,y)
double computeCoef(int m)
{
    gsl_function F;
    intparam param;
    param.beta=beta[m];
    param.J1beta=J1upp[m];
    param.Y1beta=Y1upp[m];
    F.params=&param;
    double result=0,result0=0,result1,result3,error;
    size_t neval;
    // compute oscillatory parts of the integral using a specialized weighted integration function
    size_t ncoef=100;
    gsl_integration_workspace* ws=gsl_integration_workspace_alloc(1000);
    gsl_integration_qawo_table* tbl=gsl_integration_qawo_table_alloc(beta[m], 1-lc0, GSL_INTEG_COSINE, ncoef);
    F.function=&fncAymultInitialDistribution_Osc;
    param.argadd=M_PI/4/beta[m];
    gsl_integration_qawo(&F, lc0-param.argadd, 0, 1e-10, ncoef, ws, tbl, &result1, &error);
    result1*=param.Y1beta;
    param.argadd*=3;
    gsl_integration_qawo(&F, lc0-param.argadd, 0, 1e-10, ncoef, ws, tbl, &result3, &error);
    result3*=-param.J1beta;
    gsl_integration_qawo_table_free(tbl);
    gsl_integration_workspace_free(ws);
    // compute additional non-oscillatory part by ordinary quadrature
    gsl_integration_cquad_workspace* wsc=gsl_integration_cquad_workspace_alloc(1000);
    //F.function=&fncAymultInitialDistribution;
    //gsl_integration_cquad(&F, lc0, 1, 0, 1e-12, wsc, &result, &error, &neval);  // integration of the entire function is not used
    F.function=&fncAymultInitialDistribution_nonOsc;
    gsl_integration_cquad(&F, lc0, 1, 1e-8*fabs(result1+result3), 0, wsc, &result0,&error, &neval);  // integrate only non-oscillating part
    gsl_integration_cquad_workspace_free(wsc);
    return(result0+result1+result3);
}

// the solution N(j,t) as given by Eq.24 of MM03.
double fncN(double j, double t)
{
    double result=0;
    for(int m=0; m<MMAX; m++)
        result += coef[m] * exp(-gsl_pow_2(beta[m])*D/4*t) *
            (gsl_sf_bessel_J0(beta[m]*j)*Y1upp[m] - gsl_sf_bessel_Y0(beta[m]*j)*J1upp[m]);
    return result;
}

// the expression for the flux through the loss cone boundary F(t) as given by Eq.27 of MM03
double flux(double t)
{
    double result=0;
    for(int m=0; m<MMAX; m++)
        result += beta[m]*coef[m] * exp(-gsl_pow_2(beta[m])*D/4*t) *
            (J1low[m]*Y1upp[m] - Y1low[m]*J1upp[m]);
    return result * -D*lc0/2;
}

// total flux integrated over time up to t
double fluxTotal(double t)
{
    double result=0;
    for(int m=0; m<MMAX; m++)
        result += coef[m]/beta[m] * (1-exp(-gsl_pow_2(beta[m])*D/4*t)) *
            (J1low[m]*Y1upp[m] - Y1low[m]*J1upp[m]);
    return result * -2*lc0;
}

// func to find beta[m]
double fncAroots(double x, void* params)
{
    double y=*((double*)params);
    return (gsl_sf_bessel_J0(x*y)+alpha*x*gsl_sf_bessel_J1(x*y)/2.0*y)*gsl_sf_bessel_Y1(x) - (gsl_sf_bessel_Y0(x*y)+alpha*x*gsl_sf_bessel_Y1(x*y)/2.0*y)*gsl_sf_bessel_J1(x);
}

// find roots of Bessel functions and compute expansion coefficients
void prepareAnalyticSolution()
{
    gsl_function F;
    gsl_root_fsolver* solv = gsl_root_fsolver_alloc(gsl_root_fsolver_brent); 
    J0low.assign(MMAX, 0);
    J1low.assign(MMAX, 0);
    J1upp.assign(MMAX, 0);
    Y1low.assign(MMAX, 0);
    Y1upp.assign(MMAX, 0);
    beta. assign(MMAX, 0);
    coef. assign(MMAX, 0);
    // find roots beta_m given implicitly by Eq.26 of MM03
    F.function=&fncAroots;
    F.params=&lc0;
    double betalow=0.125, betaupp=3.0, betaroot;
    while(fncAroots(betalow,&lc0)>0) betalow/=2;
    while(fncAroots(betaupp,&lc0)<0) betaupp+=0.1;
    for(int m=0; m<MMAX; m++)
    {
        int status=0, iter=0;
        gsl_root_fsolver_set (solv, &F, betalow, betaupp);
        do{
            iter++;
            status  = gsl_root_fsolver_iterate (solv);
            betaroot= gsl_root_fsolver_root (solv);
            betalow = gsl_root_fsolver_x_lower (solv);
            betaupp = gsl_root_fsolver_x_upper (solv);
            status = gsl_root_test_interval (betalow, betaupp, 0, 1e-12);
        }
        while(status == GSL_CONTINUE && iter < 100); 
        J0low[m]=gsl_sf_bessel_J0(betaroot*lc0);
        J1low[m]=gsl_sf_bessel_J1(betaroot*lc0);
        Y1low[m]=gsl_sf_bessel_Y1(betaroot*lc0);
        J1upp[m]=gsl_sf_bessel_J1(betaroot);
        Y1upp[m]=gsl_sf_bessel_Y1(betaroot);
        beta[m]=betaroot;
        betalow=betaroot+0.1;
        double sign=fncAroots(betalow, &lc0);
        betaupp=betaroot+4;
        while(fncAroots(betaupp, &lc0)*sign > 0) betaupp+=0.1;
    }

    // find the coefficients of expansion of the initial d.f. N(j,t=0) over the Bessel basis.
    // each coefficient is given by the integral \int_{j_{lc0}}^1 j' A(\beta_m,j') N(j',t=0) dj'
    // multiplied by some combination of Bessel functions -- in other words, all terms in 
    // Eq.24 of MM03 that do not depend on j and t.
    for(int m=0; m<MMAX; m++)
    {
        double result=computeCoef(m);
        coef[m] = result * gsl_pow_2(beta[m]*J0low[m]) / (gsl_pow_2(J0low[m]) - gsl_pow_2(J1upp[m])) * M_PI*M_PI/2;
    }
}

// main program begins here
int main(int argc, char* argv[])
{
    gsl_set_error_handler(&gsl_verbose_error_handler);
    Config params(argc, argv);
    alpha= params.value("alpha",0.);      // coef-t in the boundary condition
    Rmin = params.value("Rmin",-1.);      // capture boundary
    D=params.value("dc",1.);              // diffusion coefficient
    if(Rmin<=0 || Rmin>=1) { std::cerr << "Error, Rmin should be >0 and <1\n"; return 0; }
    else lc0=sqrt(Rmin);
    Rgap=params.value("Rgap",0.);         // width of the depleted region of phase space
    Pgap=params.value("Pgap",3.);         // power-law index of distribution function f(R)=(R/Rgap)^p in the gap region
    if(Rgap<=Rmin) Rgap=0;                // no gap
    if(Rgap>=1 && Pgap==0) { std::cerr << "Warning, initial conditions are identically zero\n"; return 0; }
    MMAX=params.value("mmax",MMAX);       // number of terms in the expansion
    tmax=params.value("tmax",1.);         // total interval of time to compute the solution
    tstep=params.value("tstep",tmax);     // minimum output timestep
    tcutoff=params.value("tcutoff",0.);   // time for transition from numerical to analytical solution
    if(tcutoff==0) {                      // determine the validity range of analytic solution automatically
        tcutoff = 12.0/D/gsl_pow_2(MMAX); // corresponds to exp(-30) decrease in the amplitude of the highest-m term
    }
    if(tmax>tcutoff) {
        prepareAnalyticSolution();
        while(tcutoff<tmax && (flux(tcutoff)<1e-9 || fluxTotal(tcutoff)<0))
            tcutoff*=1.5;                 // ensure that the flux is not dominated by roundoff errors
    }
    else MMAX=0;
    if(tstep<tcutoff)
        prepareNumericSolution();
    // output instantaneous capture rate and total captured mass as functions of time
    //std::cerr << "Mmax="<<MMAX<<", tstep="<<tstep<<", tcutoff="<<tcutoff<<", num.timestep="<<solverMaxStep<<"\n";
    for(double t=tstep; t<=tmax && tstep>0; t+=tstep) {
        if(t<tcutoff) solverStep(tstep);  // use numerical solver
        double fluxCur = t<tcutoff ? solverFlux() : flux(t);
        double fluxTot = t<tcutoff ? solverFluxTotal() : fluxTotal(t);
        std::cout << t << "\t" << fluxCur << "\t" << fluxTot << "\n";
        tstep*=1.1;
        if(t+tstep>tmax) tstep=tmax-t;
    }
    return 0;
}
