/** \file    common.h
    \brief   provides forward class declarations, several abstract data classes, 
             a couple of global functions, and global constants (header file only).
    \author  EV
    \date    2009-2015
*/
#pragma once
#include <utility>
#include <vector>
#include <string>

/// common namespace for all core SMILE classes, functions and variables
namespace smile{

/* ------- Global constants not changeable in runtime ------- */

/// \name ----------- Parameters for potential and force evaluation ------------- 
///@{

/// Dimensions of the real world. DO NOT CHANGE (unless you are God)!
const unsigned int N_DIM=3;

/// max number of basis-function expansion members (radial and angular). 
const size_t MAX_NCOEFS_RADIAL=101;
const size_t MAX_NCOEFS_ANGULAR=51;

/// relative accuracy of density computation
const double EPSREL_DENSITY_INT=1e-4;

/// relative accuracy of potential computation (integration tolerance parameter)
const double EPSREL_POTENTIAL_INT=1e-6;

/// absolute error in potential computation (supercedes the relative error in case of very small coefficients)
const double EPSABS_POTENTIAL_INT=1e-15;

/// relative accuracy of cell mass computation in Schwarzschild modelling
const double EPSREL_MASS_INT=1e-2;

/// inner and outer cutoff radii appearing in calculations of M(r), r(E) and related
const double MIN_RADIUS_CUTOFF=1e-6;
const double MAX_RADIUS_CUTOFF=1e+6;

/// auxiliary softening parameter for Spline potential to avoid singularities in some odd cases
const double SPLINE_MIN_RADIUS=1e-10;

///@}
/// \name ------------- Orbit integration parameters ----------
///@{

/// maximal number of integration steps (in case something went wrong with the timestep, do not dead-lock)
const size_t NUMSTEP_MAX=static_cast<size_t>(1e7);

/// for DOP853 integrator, setting this flag eliminates three calls to force function per timestep at the expense 
/// of getting a slightly lower (still acceptable) accuracy in the interpolated solution (6th instead of 7th order)
#define DOP853_DENSE_OUTPUT_FAST

/// for Nbody treecode, another safety measure is to decrease timestep near BH by this factor 
/// (in the leap-frog integrator)
const double TREECODE_TIMESTEP_BH_FACTOR=0.2;

/// whether to use quadrupole moments in tree expansion of Nbody potential (no reason to switch it off)
#define TREECODE_QUADRUPOLE 1

/// if defined, use compact Epanechnikov kernel for potential/force softening, 
/// otherwise use Plummer softening (not recommended?)
#define TREECODE_EPS_COMPACT_KERNEL 1

/// if defined, evaluate adaptive softening using true local density, 
/// otherwise use a more approximate but faster method based on cell volumes
#define TREECODE_EPS_FROM_DENSITY 0

/// if defined, use Dehnen's (2000) multipole acceptance criterion based on r_max - max distance 
/// between cell's c.o.m and its child nodes, otherwise use standard Barnes&Hut criterion based on cell size
#define TREECODE_DEHNEN_MAC 0

/// density estimate in treecode by using distance to k-th neighbour; if using spline kernel, put more (>~20), 
/// since this kernel is more concentrated
const size_t TREECODE_DENSITY_KTH_NEIGHBOUR=32;

/// method to smooth density: standard SPH spline, or Epanechnikov kernel
/// (both have finite support equal to distance to Kth neighbour)
#define TREECODE_DENSITY_SPLINE 0

/// initial value of orbit separation (if variational equation is not used and two nearby trajectories are followed)
const double LYAP_DEV_INIT=1e-10;

/// if distance between nearby trajectories reaches the following threshold, renormalize it back to LYAP_DEV_INIT
const double LYAP_DEV_MAX=1e-6;

/// if variational equation is used, perform renormalization of deviation vector if greater than threshold
/// (only for numerical convenience, since the var.equation is anyway linear)
const double LYAP_VAREQ_DEV_MAX=1e5;

///@}
/// \name ------------- Orbit spectral analysis and classification options ----------
///@{

/// orbit spectrum is calculated by fast fourier transform from GSL.
/// to avoid worst-case behavior of FFT, we ensure that its length never contains 
/// primes larger that MAX_FFT_PRIME (here equal to 42 minus unity)
const size_t MAX_FFT_PRIME=41;

/// Use Hunter's method for precise determination of frequency of a given line 
/// if the spectrum neat its maximum is reasonably similar to that of a single line. 
/// Otherwise use less precise method without Hanning filtering. Controls threshold for selection.
const double FREQ_PRECISE_THRESHOLD=5.0;

/// maximal number of spectral lines in each coordinate to search for
const size_t MAX_SPECTRAL_LINES=10;

/// do not use lines which amplitude is X times lower than amplitude of the leading line 
const double MIN_SPECTRAL_LINE_AMPLITUDE=1e-2;

/// method to estimate frequency diffusion rate.
/// 0 - difference in the largest of three frequencies (Valluri&Merritt 98), 
/// 1 - average of differences in all three freq, 
/// 2 - average over those coords for which |f1-f2|/(f1+f2)<0.5 (more robust to spurious errors in lf determination)
#define FREQ_DIFF_METHOD 2

/// accuracy of frequency comparison (to determine resonances), in units of Nyquist frequency
const double FREQ_ACCURACY=0.1;

/// minimum value of freq.diff threshold separating regular and chaotic orbits
/// (actual threshold depends on integration interval)
const double FREQ_DIFF_REG_MIN=1e-6;

/// maximum order of m:n frequency resonance (or linear dependence of three frequencies)
const int FREQ_RES_2=10;

/// maximum order of l*a+m*b+n*c frequency commensurability (thin orbit)
const int FREQ_RES_3=10;

/// threshold value of |<z>| / rmax to classify orbit as pyramid or saucer
const double OC_PYRAMID_AVGZ=0.1;

/// max number of points in Poincare section
const size_t MAX_POINCARE_POINTS=10000;

/// fraction of closest pericenter passages which are used in fitting
const double PERICENTER_FIT_FRACTION=0.1;

/// minimum number of points used in pericenter fitting
const int PERICENTER_FIT_MIN_POINTS=10;

/// defines max points in pericenter fitting
const int PERICENTER_FIT_MAX_POINTS=50;

///@}
/// \name ------------- Orbit library / Schwarzschild modelling options --------------- 
///@{
/// sample initial conditions out to a radius containing this fraction of mass
const double MAX_LAGRANGIAN_RADIUS=0.999;

/// number of radial points used in createMassModel
const size_t NUM_RADIAL_POINTS_SPHERICAL_MODEL=100; 

/// number of significant digits for initial conditions (rounded to this accuracy)
//#define NUM_DIGITS_ROUND 6  // unused

/// if using Schwarzschild stationary start space (3d), put points on a square grid 
/// in 3 sectors adjacent to X, Y and Z axes  (same as in Schwarzschild 1993).
/// otherwise, put points along lines of constant Z (=constant theta), 
/// number of points being proportional to sin(theta)  (same as in Terzic 2002).
#define SCHWARZSCHILD_START_SPACE 1

/// the mass in each cell of the classic Schwarzschild grid is computed by dividing it 
/// into a number of smaller subcells in two angular directions; this number is given below
#define NUM_SUBDIV_ANGULAR 5

/// max relative difference in cell mass to consider the cell as feasible: abs(1-Mcell/Mcellrequired)
const double SCHW_MAX_CELLMASS_REL_DIFFERENCE=1e-2;

/// min orbit weight to use the orbit after the optimization problem solved (divide by number of orbits!)
const double SCHW_MIN_ORBIT_WEIGHT_USED=1e-4;

/// Nbody export: relative importance of nullifying total angular momentum 
/// at the expense of pericenter velocity * pericenter offset
const double NBEXPORT_XV_FACTOR=10.0;

///@}
/// \name -------------- Options for GUI and high-level operations ------------
///@{

/// SMILE version identifier
#define SMILEVERSION "2.5"

/// number of points in curve representing equipotential surface (actually *4)
const size_t NUM_POINTS_PLOT_EQUIPOTENTIAL=100;

/// number of points in outer bounding curve in Poincare section (actually *4)
const size_t NUM_POINTS_PLOT_POINCARE=100;

/// min. weight of orbit to be displayed in GUI
const double SCHW_MIN_ORBIT_WEIGHT_DISPLAYED=1e-4;

/// width of frequency spectrum displayed (in units of orbital frequency)
const double PLOT_FREQ_MAX_F=5.0;

#ifdef HAVE_QWT3D
/// whether to perform triangulation and display 'envelope' of an orbit 
/// (useless without Qwt3D; requires external program QDelaunay from QHull package)
#define USE3DFACETS
#endif

/// auto-load settings from file
#define DEFAULT_SETTINGS_FILE_NAME "smile.ini"

///@}

/* ------- Declare several classes implemented in common.cpp and various other modules -------- */

// forward declarations
class CDensity;
class CPotential;
class COrbit;
class COrbitDesc;
class COrbitLibrary;
class CMassModel;
class CMassModelAxisymmetric;
class CBasicOrbitFilteringFnc;
class CBasicSchwModel;
class CBasicOrbitRuntimeFnc;
class CBasicOrbitRuntimeFncCreator;
class CBasicInformation;
template<typename NumT> struct CPosPoint;
template<typename NumT> struct CPosVelPoint;

/// \name ------- define several convenience types ------- 
///@{
typedef float sdNumType;                  ///< how to store orbit data for Schwarzschild modelling (float or double)
typedef float smNumType;                  ///< how to handle Schwarzschild modelling data internally (float or double)
typedef std::vector<double> vectord;      ///< vector of double values
typedef std::vector<float> vectorf;       ///< vector of float values
typedef std::pair<double, double> paird;  ///< a pair of doubles
typedef std::vector< paird > vectorpd;    ///< vector of pairs of doubles
typedef std::pair<float, float> pairf;    ///< a pair of floats
typedef std::vector< pairf > vectorpf;    ///< vector of pairs of floats
typedef vectord OdeStateType;             ///< container for the array of variables used in ODE integration
///@}
/// \name ----------- definition of various data structures ------------- 
///@{

/** Data-only class containing 3 coordinates and 3 velocities. 
    \tparam NumT is a numerical type (double or float).
    Copy constructor provided for both instantiations so that each one may be 
    seamlessly converted to another one by (implicit) type cast.
*/
template<typename NumT> struct CPosVelPoint
{
public:
    /// default constructor
    CPosVelPoint() { for(unsigned int i=0; i<N_DIM; i++) { Pos[i]=0; Vel[i]=0; } }
    /// construct from six float/double numbers
    CPosVelPoint( const NumT Pos0, const NumT Pos1, const NumT Pos2, const NumT Vel0, const NumT Vel1, const NumT Vel2)
    { Pos[0]=Pos0; Pos[1]=Pos1; Pos[2]=Pos2; Vel[0]=Vel0; Vel[1]=Vel1; Vel[2]=Vel2; };
    /// copy constructor from CPosVelPoint<double>
    CPosVelPoint( const CPosVelPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); Vel[i]=static_cast<NumT>(src.Vel[i]); } };
    /// copy constructor from CPosVelPoint<float>
    CPosVelPoint( const CPosVelPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); Vel[i]=static_cast<NumT>(src.Vel[i]); } };
    NumT Pos[N_DIM];   //< Components of position vector. Warning: no range check performed when setting Pos or Vel[n>=N_DIM] !
    NumT Vel[N_DIM];   //< Components of velocity vector
};

/** Data-only class containing 3 coordinates. 
    \tparam NumT may be float or double.
    Copy constructor provided for both instantiations so that each one may be 
    seamlessly converted to another one by (implicit) type cast.
*/
template<typename NumT> struct CPosPoint
{
public:
    /// default constructor
    CPosPoint() { for(unsigned int i=0; i<N_DIM; i++) { Pos[i]=0; } }
    /// construct from three float/double numbers
    CPosPoint( const NumT Pos0, const NumT Pos1, const NumT Pos2)
    { Pos[0]=Pos0; Pos[1]=Pos1; Pos[2]=Pos2; };
    /// copy constructor from CPosPoint<double>
    CPosPoint( const CPosPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosPoint<float>
    CPosPoint( const CPosPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosVelPoint<double>
    CPosPoint( const CPosVelPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosVelPoint<float>
    CPosPoint( const CPosVelPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    NumT Pos[N_DIM];   //< Components of position vector. Warning: no range check performed when setting Pos or Vel[n>=N_DIM] !
};

/** An array of particles with positions, velocities and masses.
    It is implemented as a separate structure instead of just a vector of pairs, 
    because of the limitations of C++ that don't allow to template a typedef without a class.
    \tparam NumT is a numerical type (double or float)
*/
template<typename NumT> struct CPointMassSet {
    typedef std::pair< CPosVelPoint<NumT>, NumT> ElemType;  ///< templated typedef of a single particle
    typedef std::vector< ElemType > Type;                   ///< templated typedef of an array of particles
    Type data;           ///< particles are stored in this array
    ///  default empty constructor
    CPointMassSet() {};
    /** a seamless conversion constructor from another point mass set with a possibly different template argument.
        \tparam OtherNumT is a numerical type (double or float) of the source PointMassSet */
    template<typename OtherNumT> CPointMassSet(const CPointMassSet<OtherNumT> &src) {
        data.reserve(src.size()); 
        for(size_t i=0; i<src.size(); i++) 
            data.push_back(ElemType(CPosVelPoint<NumT>(src[i].first), static_cast<NumT>(src[i].second)));
    }
    inline ElemType& at(size_t index) { return data.at(index); }             // convenience shorthand
    inline const ElemType& at(size_t index) const { return data.at(index); } // convenience shorthand
    inline ElemType& operator[](size_t index) { return data[index]; }        // convenience shorthand
    inline const ElemType& operator[](size_t index) const { return data[index]; } // another shorthand
    inline size_t size() const { return data.size(); }  // convenience shorthand
    inline void add( const CPosVelPoint<NumT> &first, const NumT second) /// convenience function to add an element
    { data.push_back(ElemType(first, second)); }
};
///@}
/// \name ----------- definition of some basic (abstract) classes to be implemented in various modules ------------- 
///@{

/** interface for matrix storage.
    used to pass around large quantities of data without making temporary copies of arrays;
    derived classes may provide virtual method which returns matrix elements, 
    or simply keep the matrix "as is" in std::vector<std::vector<NumT> >
    \tparam NumT is a numerical type (double or float)
*/
template<typename NumT> struct CMatrix
{
public:
    virtual ~CMatrix() {};
    virtual bool isDense() const { return true; }   ///< whether the matrix is dense or sparse (may be used to optimize calculations)
    virtual size_t size1() const=0;                 ///< return size for first index
    virtual size_t size2() const=0;                 ///< same for second
    virtual size_t numElems() const                 ///  number of possibly nonzero elements in sparse matrix
    { return size1()*size2(); }                     ///< by default all size1*size2 elements count, even if some of them are zero
    virtual NumT at(size_t i1, size_t i2) const=0;  ///< return value at indices i1,i2. Note: there is no abstract function to set a value.
    virtual NumT getElem(size_t index, size_t *i1, size_t *i2) const  ///  get an element from the matrix, 0<=index<numElems()
    { *i1=index/size1(); *i2=index%size1(); return at(*i1, *i2); }    ///< by default get an element from the dense matrix representation
};

/** matrix interface class that actually holds the entire matrix and returns its elements **/
template<typename NumT> struct CMatrixDense: public CMatrix<NumT>
{
public:
    CMatrixDense(size_t s1, size_t s2) 
    { matrix.resize(s1); for(size_t i1=0; i1<s1; i1++) matrix[i1].assign(s2, 0); }   //!!! what if out-of-memory exception arises here?
    virtual size_t size1() const { return matrix.size(); }
    virtual size_t size2() const { return matrix[0].size(); }
    virtual NumT at(size_t i1, size_t i2) const { return matrix.at(i1).at(i2); }  ///< return a regular value from 2d array
    void set(size_t i1, size_t i2, NumT value) { matrix.at(i1).at(i2)=value; }    ///< a way to set the value of matrix element
private:
    std::vector< std::vector<NumT> > matrix;  ///< matrix elements
};

/** matrix interface class that provides a diagonal matrix **/
template<typename NumT> struct CMatrixDiagonal: public CMatrix<NumT>
{
public:
    CMatrixDiagonal(size_t s1) 
    { diagonal.assign(s1, 0); }
    virtual bool isDense() const { return false; }
    virtual size_t size1() const { return diagonal.size(); }
    virtual size_t size2() const { return diagonal.size(); }
    virtual size_t numElems() const { return diagonal.size(); }
    virtual NumT at(size_t i1, size_t i2) const { return i1==i2 ? diagonal.at(i1) : 0; }  ///< return a value from the diagonal or zero if not on the diagonal
    virtual NumT getElem(size_t index, size_t *i1, size_t *i2) const                      ///  retrieves all values in the matrix (these are only the diagonal ones)
    { *i1=index; *i2=index; return diagonal.at(index); }
    void set(size_t i1, NumT value) { diagonal.at(i1)=value; }
private:
    std::vector<NumT> diagonal;  ///< diagonal matrix elements
};

/** The interface class for various third-party quadratic optimization solvers.

    solves the linear or quadratic optimization problem defined as the linear system 
      M sol = rhs, 
    where M is N_v * N_c matrix, sol is N_v vector of variables to be found, 
    rhs is N_c vector of constraints to be satisfied;
    the solution vector has to be non-negative and <= maxSolValue if the latter is provided.

    Of all possible solutions, the one is chosen that minimizes the objective (cost) function
      CF = 1/2 sol^T Q sol + L^T sol, 
    where Q is N_v*N_v matrix (most likely a sparse one), and L is N_v vector.

    If Q=NULL then the problem is solved by the linear programming method, 
    otherwise the quadratic programming is employed.

    Return value is 0 for success, <0 for error code;
    standard error codes are:
    -1  model is infeasible
    -2  not enough memory

    \tparam NumT is a numerical type (double or float)
*/
template<typename NumT> class CBasicQuadraticOptimizationSolver
{
public:
    virtual ~CBasicQuadraticOptimizationSolver() {};
    virtual int callSolver(
        const CMatrix<NumT> *linearMatrix,      ///< matrix M of linear equations for the optimization problem  "M w = rhs". Passed as an interface to an abstract class, must not be NULL
        const std::vector<NumT> &rhs,           ///< constraints to be satisfied
        const std::vector<NumT> &maxSolValue,   ///< upper limits for sol[v] (zero means no limit)
        const std::vector<NumT> &objFuncLinear, ///< vector L in the objective function, which is multiplied by the solution vector
        const CMatrix<NumT> *objFuncQuadratic,  ///< matrix Q in the objective function, may be NULL
        std::vector<NumT> *sol                  ///< returns solution in this vector
    ) const=0;                                  ///< \returns 0 for success, otherwise an error code
    virtual std::string errorDescription(int result) const=0;  ///< reports text information for a given error code
};

/** Parent class for orbit filtering functions which evaluate whether an orbit should be used (in some calculation).
    Return value is a float number (for example, between 0 and 1 for telling 
    how well an orbit satisfies some criterion, but may be >1 or negative).

    Example of the usage of these functions is in COrbitDesc::getOrbitPopulation() routine 
    which computes statistics based on acceptance criteria returned by such a function, 
    or in assigning penalty to orbits of specified sort in Schwarzschild modelling.
*/
class CBasicOrbitFilteringFnc
{
public:
    virtual ~CBasicOrbitFilteringFnc() {};
    virtual double eval(const COrbitDesc* orbit) const = 0;
};

/** defines a routine which may be called after each integration timestep 
    to perform user-specific data collection.
    (such as computing cell occupation times in Schw model or recording Poincare section points).
*/
class CBasicOrbitRuntimeFnc
{
public:
    /// lists all possible runtime functions
enum FNCTYPE {
    FT_TRAJ_ANALYSIS,   ///< COrbitRuntimeTrajectory
    FT_TRAJ_SAMPLE,     ///< COrbitRuntimeTrajSample
    FT_POINCARE,        ///< COrbitRuntimePoincare
    FT_PERICENTER,      ///< COrbitRuntimePericenter
    FT_SCHW,            ///< COrbitRuntimeSchw
    FT_VEL_PERTURBER    ///< COrbitRuntimeVelPerturber
};
    /// result of processing of a timestep 
enum STEPRESULT {
    SR_CONTINUE,        ///< need to continue integration
    SR_TERMINATE,       ///< may terminate integration if necessary
    SR_REINIT           ///< variables have changed, need to re-initialize integrator
};
    /// constructs instance and initializes some internal data
    explicit CBasicOrbitRuntimeFnc(const COrbit* _orbit): orbit(_orbit) {} ;

    /// destroys all necessary internal data
    virtual ~CBasicOrbitRuntimeFnc() {};

    /// derived classes return their type
    virtual FNCTYPE FncType() const=0;

    /** the following method is called after each timestep.
        \param[in] told is the value of time at the beginning of the timestep
        \param[in] tcur is the time at the end of timestep
        \param[in,out] vars are the phase-space variables at the end of timestep; 
        the function may change them and return SR_REINIT to indicate that the ODE integrator 
        needs to be restarted.

        This procedure may call orbit->getPos() and orbit->getVel()
        to get interpolated coordinates on the interval [told, tcur].
        \return true if integration may be finished right now [default], false if needs to continue.
        (the ultimate decision is taken by orbit integration routine and may ignore the demands of runtime functions) 
    */
    virtual STEPRESULT Timestep(const double told, const double tcur, CPosVelPoint<double> *vars)=0;

    /// called when integration is finished, performs cleanup/postprocessing but doesn't destroy collected data
    virtual void Finish() {};

    /// returns the collected data as an instance of an information object
    virtual CBasicInformation* getData() const = 0;
protected:
    const COrbit* orbit;  ///< pointer to the orbit that is being integrated
};

/** defines a basic class that creates some sort of runtime function for any orbit in the orbit library.
    Derived classes may receive additional data in constructor and emit specific types 
    of timestep functions for a given orbit.
    Thus the interface between orbit library integrator and the specific class of runtime functions 
    is mediated by corresponding creator classes, one for each class of orbit runtime fnc.
*/
class CBasicOrbitRuntimeFncCreator
{
public:
    virtual ~CBasicOrbitRuntimeFncCreator() {};

    /// create an instance of the runtime function for a given orbit.
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const=0;
};

/** parent class for data container objects used to store various data and exchange it between objects.
    Data stored here is related to some orbit properties (e.g. orbit classification results, 
    or points in the trajectory, or data for Schwarzschild modelling).
    The instances of runtime functions derived from CBasicOrbitRuntimeFnc return specific information 
    objects which contain data collected about the orbit during integration.
    Classes defined in fileio.h handle storing and loading data.
    Classes derived from CBasicSchwData also use the CSchwInformation class to store data for modelling.
*/
class CBasicInformation
{
public:
    /// lists all possible information types
enum INFOTYPE {
    IT_UNKNOWN,         ///< dark matter
    IT_TRAJ_ANALYSIS,   ///< orbit analysis: COrbitInformation
    IT_TRAJ_SAMPLE,     ///< CTrajSampleInformation
    IT_POINCARE,        ///< CPoincareInformation [unused]
    IT_PERICENTER,      ///< CPericenterInformation
    IT_SCHW,            ///< CSchwInformation
    IT_VEL_PERTURBER    ///< CVelPerturberInformation
};
    virtual ~CBasicInformation() {};
    virtual INFOTYPE infoType() const=0;  ///< return the derived class type
    /// present a textual representation of the data contained in this class (may be empty string)
    virtual std::string toString() const = 0;
};

/** basic class of ODE system used in the orbit integrator. 
    It provides a function for computing the right-hand side of the system of differential equations, 
    and a number of auxiliary functions (for initializing the system and providing some operations).
    The COrbit class is descending from this one.
*/
class COdeSystem {
public:
    /// result of processing of a timestep 
enum STEPRESULT {
    SR_CONTINUE,        ///< need to continue integration
    SR_TERMINATE,       ///< may terminate integration if necessary
    SR_REINIT           ///< variables have changed, need to re-initialize integrator
};
    COdeSystem() {};
    virtual ~COdeSystem() {};
    virtual void initState(OdeStateType* state)=0;
    virtual void odeFnc(double t, const OdeStateType& x, OdeStateType* dx)=0;
    virtual bool isStdHamiltonian() { return false; };
    virtual double getRelErrorToleranceFactor(const OdeStateType& /*state*/) const { return 1; };
    virtual STEPRESULT processStep(double timePrev, double timeCurr, OdeStateType* state)=0;
};

/** basic class for numerical integrators of ODE system */
class CBasicOdeIntegrator
{
public:
    /// list of various ODE stepper types
enum STEPPERKIND {
    SK_DEFAULT,
    SK_LEAPFROG_NB,
    SK_DOP853,
    SK_IAS15,
    SK_HERMITE,
    SK_ODEINT_CK5,
    SK_ODEINT_DP5,
    SK_ODEINT_BS3,
    SK_ODEINT_BS,
    SK_ODEINT_RK4,
    SK_ODEINT_SYMPL4
};

    CBasicOdeIntegrator(COdeSystem* _odeSystem):
      odeSystem(_odeSystem), timePrev(0), timeCurr(0) {};

    virtual ~CBasicOdeIntegrator() {};

    /** perform integration */
    virtual void integrateToTime(double timeEnd)=0;

    /** return the time to which the integration has proceeded so far */
    double getTime() const { return timeCurr; }

    /** return interpolated value of c-th variable at time t, where t must lie within current timestep interval;
        called from various runtime functions that need to obtain x(t) at arbitrary times */
    virtual double getInterpolatedSolution(unsigned int c, double t) const=0;

protected:
    COdeSystem* odeSystem;
    double timePrev, timeCurr;      //< previous and current value of time to which the integration has advanced
};

///@}
/// \name -------- Various convenience definitions --------
///@{

/// convenience definition of vectors of timestep functions and function creators
typedef std::vector<CBasicOrbitRuntimeFnc*> vectorRuntimeFncs;
typedef std::vector<const CBasicOrbitRuntimeFncCreator*> vectorRuntimeFncCreators;
typedef std::vector<const CBasicInformation*> vectorInformation;

/// convenience definitions of templated vectors containing particle sets with/without masses
typedef std::vector< CPosVelPoint<double> > CPosVelDataDouble;
typedef std::vector< CPosVelPoint<float > > CPosVelDataFloat;

/// convenience shorthand for squaring a number
inline double pow_2(double x) { return x*x; }
inline float pow_2(float x) { return x*x; }
inline size_t pow_2(size_t x) { return x*x; }

/// convenience function for sorting arrays of std::pair<float/double, something>
template<typename T1, typename T2> bool comparepair(const std::pair<T1,T2>& val1, const std::pair<T1,T2>& val2)
{  return val1.first < val2.first;  }

///@}

// below is the text that appears on the main doxy page //
/**
\mainpage SMILE documentation

\section sec_intro  Introduction

SMILE is a software for orbit analysis and Schwarzschild modelling of stellar systems.
The primary applications are:
- Exploring orbits in various potentials, either given by analytical formulae 
  or several approximate expansions;
- Studying properties of orbits, computed by the builtin orbit integrator 
  or obtained from external data;
- Constructing self-consistent equilibrium models of triaxial stellar systems 
  with a given density profile using the Schwarzschild method.

The software comes in two versions -- the GUI interactive program 
(implemented in CSmileGUI class), which forms an interface over the main 
business logic layer implemented in CSmileCore class,
and a console tool whose functionality is implemented in CSmileConsole class.
In addition, there are several standalone utility programs:
- mkspherical.cpp (create spherical mass models from a text table or from N-body snapshots)
- renderdensity.cpp (create a visual representation of a non-spherical density model specified 
  by a BSE/Spline potential)
- measureshape.cpp (measure the axis ratio of an N-body snapshot as a function of radius)
- testaccuracy.cpp (check the accuracy of potential approximation constructed from an N-body snapshot)
- snaporbits.cpp (perform orbit analysis on trajectories of particles from an N-body snapshot)
- energydiff.cpp (measure the rate of energy diffusion in N-body simulation)

The user's guide and the mathematical background behind several algorithms employed 
in the software can be found <a href="../readme.pdf">here</a>, while the scientific description 
of Schwarzschild modelling is given in the <a href="../codepaper.pdf">code paper</a>.

\section sec_compile  Obtaining and compiling the software

The software can be obtained at http://td.lpi.ru/~eugvas/smile/.
Compiled versions for several platforms are available; if you need to compile it from source 
then read this section.

The following third-party libraries are required:
- Qt framework (tested with version 4.8):  http://qt-project.org/
- GNU Scientific Library:  http://www.gnu.org/software/gsl/
- Qwt plotting library (for GUI only):  http://sourceforge.net/projects/qwt/files/qwt/5.2.1/

The following libraries or external programs are optional:
- interp2d library:  https://github.com/diazona/interp2d/
- GLPK (linear-programming kit):  http://www.gnu.org/software/glpk/
- CVXOPT (Python-based linear/quadratic optimization solver):  http://www.cvxopt.org/
- BPMPD (standalone linear/quadratic solver):  http://www.sztaki.hu/~meszaros/bpmpd/
  (of the three options above, at least one is required for the Schwarzschild modelling to work).
- QwtPlot3d library:  http://qwtplot3d.sourceforge.net/
- qdelaunay program from QHull package:  http://www.qhull.org/download/
- UNSIO library for N-body snapshot manipulation:  http://projets.oamp.fr/projects/unsio
- HDF5 library for binary data storage in structured files:  http://www.hdfgroup.org/HDF5/
- boost::numeric::odeint for several additional ODE integrators

Check smile.pri for the correct paths to libraries and set the compilation options 
HAVE_*** according to the available optional components, then run 
    `qmake smile.pro`, then `make`; repeat for `qmake smilec.pro`.


\section sec_history  Version history

v1.0 (March 2011): initial release (not advertised anywhere)

v1.1 (July 2011): implemented a broader class of basis-set expansion, fixed some bugs.

v2.0 (1 August 2013): first major public release, accompanies the paper [Vasiliev, 2013, MNRAS]
- program structure substantially reworked towards modularity;
  it should be possible to use only potential or orbit integrator/analysis in 
  other programs by including a subset of source files.
- New class of general-purpose potential approximation (Spline spherical-harmonic expansion).
- Choice of symmetry degree for BSE/Spline (from 'None' through standard 'Triaxial' to 'Spherical').
- Schwarzschild modelling: separated solver (several possible variants) from 
  data objects and the model itself; 
- different choices for Schwarzschild model type (classical grid-based, basis set 
  and spherical harmonic expansion). 
- Different method of generating initial conditions for Schwarzschild model 
  (based on distribution function for an equivalent spherical model, constructed 
  by Eddington inversion formula in CMassModel class).
- A number of features were scrapped: initial conditions for SM are only generated randomly,
  no stationary/principal-plane start space anymore (it did not produce good models anyway),
  no 'orbit bunches' (or 'dithering', which averages over several nearby orbits: 
  just increase the total number of orbits, and quadratic optimization will drive the orbit 
  weights towards more uniformity). 
  Lucy iterations optimization removed -- was too slow on realistic problem sizes and never 
  outperformed quadratic optimization. 

v2.1 (late 2013): non-public, accompanies the paper [Vasiliev,Antonini,Merritt, 2014, ApJ]
- added an option for initial conditions refinement at the center.
- added Schwarzschild data object for constraining the relative angular momentum distribution
  at a given energy.

v2.2 (March 2014): non-public, accompanies the papers [Vasiliev, 2014, Clas.Quant.Grav. and Vasiliev, 2014, MNRAS]
- several new orbit integrators using `odeint` library from `boost`.
- added velocity perturbations (for the non-spherical Monte-Carlo code `RAGA`).
- improved robustness of handling of spherical mass models.

v2.5 (1 February 2015): second major public release, accompanies the paper [Vasiliev&Athanassoula 2015]
substantially expanded feature set:
- added figure rotation to the potential.
- added several new potential and density models: Ferrers, Miyamoto-Nagai, Sersic, ExpDisk;
  potential expansion for finite-extent models based on spherical harmonics and Bessel functions;
  cylindrical spline potential expansion for highly flattened models; 
- new variants of Schwarzschild density data object for cylindrical geometry.
- reworked the interface for initial conditions generation: 
  better sampling of density profile (based on self-tuning Monte-Carlo sampling), 
  spherical and axisymmetric Jeans models for velocity sampling.
- support for multi-component potential and Schwarzschild models.
- new orbit integrators: IAS15 and Hermite.
- characteristic diagram of periodic orbits (presently not used for generating initial conditions), 
  routine for locating periodic orbits in the Poincare surface of section.
- SMILEPOT library exposing the potential machinery for external applications.

\section sec_overview  Overview of the software structure

The class hierarchy of SMILE is designed to have reasonably abstract interfaces between 
different parts of the code. 
- Classes derived from CDensity and CPotential are used to represent a density model 
  or a density-potential pair; in the latter case functions are provided to compute potential 
  and forces at a given point, which are necessary for orbit integration.
  These classes are 'standalone' and may be used in external software without carrying all 
  the dependencies from the entire SMILE machinery.
- COrbit is an interface class between orbit integrators (derived from CBasicOdeIntegrator) 
  and orbit analysis routines (derived from CBasicOrbitRuntimeFnc), 
  the exact nature of their interaction is described below. 
  It contains the initial data for the orbit (COrbitInitData) and the results of integration 
  and classification.
- The orbit integrators are advancing the solution with their internal, automatically chosen 
  timestep; after each timestep they call one or more orbit runtime functions that are 
  attached to the orbit. These functions, in turn, may obtain position and velocity at any 
  moment of time within the given timestep (for example, at regular intervals of time 
  independent of the actual integration timestep).
- The runtime functions (CBasicOrbitRuntimeFnc) thus serve dual purpose: they collect data 
  during the orbit integration, and then may perform some kind of analysis upon completion.
  The results of this analysis are passed around as instances of objects derived from 
  CBasicInformation (generally, each runtime function class has a corresponding information 
  class). 
  The runtime functions attached to COrbit persist after finishing the integration, as long 
  as the COrbit object exists. (This makes sense only in the GUI for studying a single orbit, 
  but not in the context of creating the orbit library).
- The abstractly defined data in the array of CBasicInformation objects is attached to an orbit 
  after the integration, and may be stored or loaded from disk by the I/O routines in iodata.h
- COrbitDesc is another intermediate class that represents the orbit initial data and 
  the same array of information objects as COrbit itself. Internally, it creates an instance 
  of COrbit class to do the integration and analysis; however, it deletes this object 
  upon finishing the integration and thus not retain runtime functions attached to COrbit;
  only the information resulting from analysis is kept.
- COrbitLibrary is a collection of COrbitDesc objects, and functions to assign initial 
  conditions to them. Before the integration of the library, only the initial conditions are 
  assigned; after completion each COrbitDesc keeps the analysis information.
- Another intermediate class family is derived from CBasicOrbitRuntimeFncCreator. 
  Each class of COrbitRuntimeFnc has its own creator class, whose sole purpose is to create 
  an instance of runtime function for each orbit. An array of creator classes is passed to 
  the COrbitLibrary and down to COrbitDesc and COrbit, so each orbit creates the array of 
  runtime functions for itself. (Such an intermediate step is required because runtime 
  functions know only about the particular orbit, not the entire orbit library).
- The classes for performing the Schwarzschild modelling are also split into several groups.
  There are classes for dealing with various Schwarzschild Data, classes for performing 
  the modelling, and interface classes for various third-party optimization solvers.
  Let's consider them in turn.
  We define the process of Schwarzschild modelling as matching the source data (represented
  by the orbit library and the associated orbit analysis information) to the target model, 
  which consists of one or more data objects describing the desired properties of the galaxy.
- Abstract CBasicSchwData class defines a general way of providing the required properties 
  of the model, e.g. the density distribution parametrized as the masses of cells of spatial 
  grid in the classic Schwarzschild model, or some sort of kinematic data coming from 
  the observations (although this is not implemented in the current version).
  The data objects may take the potential/density class as the input parameter and compute 
  the required quantities, which we refer to as constraints.
  In addition, they provide methods to compute the same quantities for each orbit. 
  This is mediated by a corresponding orbit runtime function; for each kind of Schwarzschild 
  data class there is a runtime function that computes the necessary data for an orbit 
  and stores it in the CSchwInformation class in the form of a one-dimensional array.
- The modelling classes, derived from CBasicSchwModel, take the array of CBasicSchwData 
  objects as the target, and the COrbitLibrary, in which all orbits have necessary kind 
  of CSchwInformation for every CBasicSchwData object in the model, as the source. 
  The modelling itself amounts to fitting the source to the target, i.e. finding the orbit 
  weights that minimize or nullify deviations of constraints combined from all orbits 
  from their required values given in schw.data objects. The precise method of finding 
  the solution, and additional conditions imposed on the distribution of orbit weights, 
  are left for specification in the derived classes. In the present version, there is 
  only one class, CSchwModelQuadOpt, which uses the linear or quadratic optimization 
  as the way of finding the solution. The task of the modelling object is to combine 
  all data objects and orbits into a single matrix equation to be passed to the solver.
  Meanwhile, to save memory by not copying large amount of data in matrices, 
  this matrix equation is represented as an abstract matrix interface (derived from 
  CMatrix) rather than an actual two-dimensional array of numbers.
  Additional preferences about what orbits should have more weight in the model 
  are provided by a CBasicOrbitFilteringFnc (see below).
- The quadratic optimization solver used in the modelling is represented by an instance 
  of a class derived from CBasicQuadraticOptimizationSolver. This class has a single 
  method to call the solver for a given matrix equation, which is taken from the 
  modelling object via CMatrix interface. The derived classes present a unified 
  interface to a particular third-party linear or quadratic solver used as the back-end 
  number cruncher.
- Of the classes that we have not considered so far, there are several simple ones that 
  represent single point masses and point mass arrays (CPointMassSet). 
  They are used to read and save Nbody snapshots. 
  Reading snapshots is used to construct CPotential objects of one of the general-purpose 
  potential solvers from a mass model specified by a discrete point mass set. 
  Writing snapshots is used in converting the Schwarzschild model to an Nbody model, 
  which is performed by COrbitLibrary.
- Reading and writing Nbody snapshots is done through classes derived from 
  CBasicIOSnapshot, which handle the representation of snapshots in various formats 
  (text, NEMO, GADGET).
- Classes derived from CBasicOrbitFilteringFnc provide a general interface for 
  evaluating an orbit according to some criterion and obtain a numerical value for it. 
  One example is the weighting of orbits in the Schwarzschild model; 
  another is used in the context of generation of the Nbody model with unequal masses 
  of particles (mass refinement). 
  The refinement strategy is provided by a CMassRefinementFnc.
- Input/output of the orbit library and all data associated with orbits is provided 
  by a pair of classes, CFileInputText / CFileOutputText. 
- Finally, to put everything together, there is CSmileCore class that provides all 
  basic functionality for both GUI and console versions of the program. 
  This is (almost) the only place where Qt framework is used, for creating helper 
  threads to perform various tasks in parallel with the main program, passing messages 
  between threads, handling the configuration information stored in INI files, etc. 
- The access to the CSmileCore object is provided by CSmileGUI or CSmileConsole, 
  depending on the variant of the main program.

Ufff.. that's the basic outline of the relationships between various classes in SMILE;
for details refer to the documentation of particular modules and classes.
It looks quite complex, but probably is inevitable to achieve the necessary level 
of generality and abstraction.

\section sec_code  Coding conventions

- The program makes extensive use of C++ concepts such as classes, templates, STL algorithms, etc. 
  However, some parts that interface with C code are inevitable (in particular, GSL routines often 
  require wrappers to deal with data structures belonging to object instances; also the old C code 
  is used in the DOP853 orbit integrator). 
  All memory allocation is done in terms of `new`/`delete` operators or by `std::vector`s; 
  we don't use C-style `malloc`/`free` or char* strings, nor the more advanced "smart pointers" 
  (or anything else from `boost`, except for the optional `odeint` library).
- Exceptions are not used consistently throughout the software. Most often, routines that may 
  generate an out-of-memory or other notable exception, will wrap the internal code in a try-catch 
  block and will safely return an error code instead of passing out the exception.
- There are several class hierarchies derived from abstract classes (named CBasic***).
  Often the exact nature of the instance of a derived class needs to be known by a non-class 
  function; in these cases, the class type is provided by some sort of getType() member functions;
  these return a enum value from a pre-defined list, which is declared in the base class. 
  Thus we do not rely on RTTI to obtain the typename of a class, but rather use this contrived 
  mechanism that needs to be manually extended in the case that a new derived class is implemented.
- const correctness is maintained throughout the code.
  Member functions that do not change the internal state of the object have `const` qualifiers.
  Input parameters are labelled as `const myType& myVar` or `const myType* myVarPtr` in the case 
  that `myType` is not a simple type. Output parameters are passed as `myType* outVar` regardless 
  of whether `myType` is a simple or complex type. (Of course, if there is only one output 
  variable, it is returned as the return type).
- On several occasions, numerical data may be provided both in `float` or `double` variants. 
  For these cases, we use template classes with template parameter NumT; often they have 
  templated constructors to allow a seamless conversion from a different data type.

*/

}  // namespace

