/** \file    icgenerator.h
    \brief   Classes for generating the initial conditions for mkspherical (stripped-down file from SMILE)
    \author  EV
    \date    2009-2014

*/

#pragma once
#include "common.h"

namespace smile{

/** Base class for creating initial conditions for orbit library.
    The output consists of starting points and optionally the priors on orbit weights 
    (if used for Schwarzschild modelling) stored in a point mass set.
*/
class CBasicICGenerator
{
public:
/// list all possible types of descendant classes
enum ICTYPE {
    IC_E_STATIONARY,     ///< CICGeneratorStationary
    IC_E_PRINCIPALPLANE, ///< CICGeneratorPrincipalPlane
    IC_E_YALPHA,         ///< CICGeneratorYalpha
    IC_E_RANDOM,         ///< CICGeneratorRandomE
    IC_DENS_SAMPLER,     ///< CICGeneratorDensitySampler
    IC_DENS_ELLIPSOIDAL, ///< CICGeneratorDensityEllipsoidal
    IC_VEL_EDDINGTON,    ///< CICGeneratorVelocityEddington
    IC_VEL_JEANSSPH,     ///< CICGeneratorVelocityJeansSph
    IC_VEL_JEANSAXI      ///< CICGeneratorVelocityJeansAxi
};
    virtual ~CBasicICGenerator() {};
    /** Create a desired number of points and add them to the point mass set. 
        \param[in] numPoints is the requested number of points; the actual output may slightly differ
        \param[in,out] points is the pointer to the array that will contain the points.
        For the IC velocity generators, this array should already contain positions of particles,
        and only the velocities will be initialized. */
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points)=0;
};

/** sample points from a spherical mass model */
class CICGeneratorDensitySpherical: public CBasicICGenerator
{
public:
    CICGeneratorDensitySpherical(const CMassModel* _model, int _quiet) :
      CBasicICGenerator(), model(_model), quiet(_quiet) {};
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CMassModel* model;
    /// controls the regularity of sampling of radius: 
    /// 0 - random sampling; 2 - quiet: place points at successive radii enclosing equal amount of mass;
    /// 1 - intermediate: place points randomly inside successive intervals enclosing equal amount of mass.
    int quiet;
};

/** sample velocities from the Eddington distribution function for equivalent spherical system */
class CICGeneratorVelocityEddington: public CBasicICGenerator
{
public:
    /// construct the instance of spherical mass model for a given density model in an external potential;
    /// if the density and potential are the same, only the latter is provided
    CICGeneratorVelocityEddington(const CPotential* _potential, const CDensity* _density=NULL, double _weightSkewFactorL=0) :
      CBasicICGenerator(), potential(_potential), density(_density), weightSkewFactorL(_weightSkewFactorL) {}; 
    virtual void createInitialConditions(size_t numPoints, CPointMassSet<float> *points);
private:
    const CPotential* potential;  ///< total potential of the system
    const CDensity* density;      ///< density of the component considered, or NULL if entire potential is used
    double weightSkewFactorL;     ///< biasing factor for increasing the sampling frequency of low angular momentum orbits
};

}; // namespace
